"use strict";
/**
 * Copyright 2024 Angus.Fenying <fenying@litert.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.createSyncFilterManager = exports.createAsyncFilterManager = void 0;
const E = require("./Errors");
class AbstractFilterManager {
    constructor() {
        this._filters = {};
    }
    register(opts) {
        opts.priority ?? (opts.priority = 0);
        if (!this._filters[opts.name]) {
            this._filters[opts.name] = [];
        }
        if (this._filters[opts.name].find((v) => v.key === opts.key)) {
            throw new E.E_DUP_FILTER_FUNCTION({ metadata: { name: opts.name, key: opts.key } });
        }
        this._filters[opts.name].push({
            key: opts.key,
            fn: opts.callback,
            priority: opts.priority
        });
        this._filters[opts.name] = this._filters[opts.name].sort((a, b) => a.priority - b.priority);
        return this;
    }
    unregister(name, key) {
        if (!this._filters[name]) {
            return this;
        }
        if (undefined === key) {
            delete this._filters[name];
            return this;
        }
        const index = this._filters[name].findIndex((v) => v.key === key);
        if (index !== -1) {
            this._filters[name].splice(index, 1);
        }
        return this;
    }
}
class AsyncFilterManager extends AbstractFilterManager {
    async filter(name, value, ...args) {
        const items = this._filters[name];
        if (!items) {
            return value;
        }
        for (const filter of items) {
            const v = filter.fn(value, ...args);
            // eslint-disable-next-line @typescript-eslint/no-confusing-void-expression
            value = v instanceof Promise ? await v : v;
        }
        return value;
    }
}
class SyncFilterManager extends AbstractFilterManager {
    filter(name, value, ...args) {
        const items = this._filters[name];
        if (!items) {
            return value;
        }
        for (const filter of items) {
            value = filter.fn(value, ...args);
        }
        return value;
    }
}
function createAsyncFilterManager() {
    return new AsyncFilterManager();
}
exports.createAsyncFilterManager = createAsyncFilterManager;
function createSyncFilterManager() {
    return new SyncFilterManager();
}
exports.createSyncFilterManager = createSyncFilterManager;
//# sourceMappingURL=FilterManager.js.map