"use strict";
/**
 * Copyright 2024 Angus.Fenying <fenying@litert.org>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   https://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.AbstractHttp2Client = void 0;
const C = require("../Common");
const AbstractProtocolClient_1 = require("./AbstractProtocolClient");
const $H2 = require("http2");
const stream_1 = require("stream");
const INITIAL_CONN_ID_COUNTER = 0;
class AbstractHttp2Client extends AbstractProtocolClient_1.AbstractProtocolClient {
    constructor(_) {
        super();
        this._ = _;
        this._connIndex = INITIAL_CONN_ID_COUNTER;
        this._connections = {};
    }
    close() {
        for (const k in this._connections) {
            const pool = this._connections[k];
            for (const connId in pool.connections) {
                pool.connections[connId].connection.destroy();
            }
        }
    }
    /**
     * Preprocess the headers of request, make all the header-names lowercase, and replace
     * H2 headers into H1 headers.
     */
    _preprocessHeaders(headers) {
        const ret = {};
        for (const k in headers) {
            const key = k.toLowerCase();
            switch (key) {
                case $H2.constants.HTTP2_HEADER_CONTENT_LENGTH:
                    ret[C.Headers.CONTENT_LENGTH_H1] = headers[k];
                    break;
                case $H2.constants.HTTP2_HEADER_METHOD:
                case $H2.constants.HTTP2_HEADER_AUTHORITY:
                    continue;
                default:
                    ret[key] = headers[k];
            }
        }
        return ret;
    }
    async _getConnection(opts, h2Opts, key) {
        let pool = this._connections[key];
        if (pool) {
            if (pool.quantity) {
                for (const k in pool.connections) {
                    const conn = pool.connections[k];
                    if (conn.concurrency < opts.concurrency && !conn.connection.closed) {
                        conn.concurrency++;
                        return [k, conn];
                    }
                }
            }
        }
        else {
            pool = this._connections[key] = {
                'connections': {},
                'quantity': 0,
                'maximum': opts.maxConnections
            };
        }
        return new Promise((resolve, reject) => {
            const session = $H2.connect(this._.getAuthority(opts.url), h2Opts);
            session.once('connect', () => {
                const connId = this._connIndex++;
                pool.quantity++;
                pool.connections[connId] = {
                    concurrency: 1,
                    connection: session
                };
                session.on('close', () => {
                    pool.quantity--;
                    delete pool.connections[connId];
                    if (!pool.quantity) {
                        delete this._connections[key];
                    }
                });
                session.removeAllListeners('error');
                resolve([connId, pool.connections[connId]]);
            }).once('error', reject);
        });
    }
    _releaseConnection(key, connId, conn) {
        conn.concurrency--;
        if (conn.connection.closed) {
            const pool = this._connections[key];
            if (pool) {
                delete pool.connections[connId];
                pool.quantity--;
                if (!pool.quantity) {
                    delete this._connections[key];
                }
            }
        }
    }
    async _processRequest(opts, key = this.getAuthorityKey(opts)) {
        const headers = {
            ...this._preprocessHeaders(opts.headers),
            [$H2.constants.HTTP2_HEADER_METHOD]: opts.method,
            [$H2.constants.HTTP2_HEADER_PATH]: this._.buildPath(opts.url)
        };
        if (!headers[$H2.constants.HTTP2_HEADER_AUTHORITY]) {
            headers[$H2.constants.HTTP2_HEADER_AUTHORITY] = opts.url.hostname;
        }
        if (opts.connectionOptions.remoteHost) {
            opts.connectionOptions.servername = opts.url.hostname;
            opts.url.hostname = opts.connectionOptions.remoteHost;
        }
        const [connId, conn] = await this._getConnection(opts, this._prepareOptions(opts), key);
        try {
            const REQ_ENTITY = this._.requireEntity(opts.method);
            if (REQ_ENTITY) {
                /**
                 * Here will validate the type of entity.
                 */
                this._preprocessEntity(opts);
                headers[$H2.constants.HTTP2_HEADER_CONTENT_LENGTH] = opts.headers[C.Headers.CONTENT_LENGTH_H1];
            }
            const req = conn.connection.request(headers, opts.requestOptions);
            if (REQ_ENTITY) {
                if (opts.data instanceof stream_1.Readable) {
                    opts.data.pipe(req);
                }
                else if (req.writable) { // if DELETE/GET/HEAD/OPTIONS/TRACE, the writable will be false.
                    req.end(opts.data);
                }
                else {
                    req.end();
                }
                delete opts.data;
            }
            else {
                req.end();
            }
            if (opts.timeout) {
                req.setTimeout(opts.timeout);
            }
            return await new Promise((resolve, reject) => {
                req.on('response', (respHeaders) => {
                    resolve({
                        'protocol': opts.connectionOptions.createConnection ? C.EProtocol.HTTPS_2 : C.EProtocol.HTTP_2,
                        'gzip': opts.gzip,
                        'deflate': opts.deflate,
                        'stream': req,
                        'headers': respHeaders,
                        'statusCode': parseInt(respHeaders[$H2.constants.HTTP2_HEADER_STATUS]),
                        'contentLength': respHeaders[$H2.constants.HTTP2_HEADER_CONTENT_LENGTH] === undefined ?
                            Infinity : parseInt(respHeaders[$H2.constants.HTTP2_HEADER_CONTENT_LENGTH]),
                        'noEntity': !this._.hasEntity(opts.method),
                    });
                })
                    .once('error', (e) => {
                    req.removeAllListeners('error');
                    req.removeAllListeners('response');
                    req.removeAllListeners('close');
                    reject(e);
                })
                    .on('close', () => { this._releaseConnection(key, connId, conn); });
            });
        }
        catch (e) {
            this._releaseConnection(key, connId, conn);
            throw e;
        }
    }
}
exports.AbstractHttp2Client = AbstractHttp2Client;
//# sourceMappingURL=AbstractHttp2Client.js.map